<?php

declare(strict_types=1);

namespace Drupal\image_sizes;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Image\ImageFactory;
use Drupal\file\FileInterface;
use Drupal\image\Entity\ImageStyle;
use Drupal\image\Plugin\Field\FieldType\ImageItem;
use Drupal\responsive_image\ResponsiveImageStyleInterface;

/**
 * Service for sizing images.
 *
 * @package Drupal\image_sizes
 */
class ImageSizesService implements ImageSizesServiceInterface {

  /**
   * Constructs the ImageSizesService service.
   *
   * @param \Drupal\Core\File\FileSystemInterface $fileSystem
   *   The file system service.
   * @param \Drupal\Core\Image\ImageFactory $imageFactory
   *   The image factory.
   *   The entity type manager service.
   * @param \Drupal\Core\Extension\ModuleHandlerInterface $moduleHandler
   *   The module handler.
   * @param \Drupal\Core\File\FileUrlGeneratorInterface $fileUrlGenerator
   *   The file URL generator.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entityTypeManager
   *   The entity type manager.
   */
  public function __construct(
    protected readonly FileSystemInterface $fileSystem,
    protected readonly ImageFactory $imageFactory,
    protected readonly ModuleHandlerInterface $moduleHandler,
    protected readonly FileUrlGeneratorInterface $fileUrlGenerator,
    protected readonly EntityTypeManagerInterface $entityTypeManager
  ) {}

  /**
   * Builds attributes based on a given file and responsive_image_style.
   *
   * @param \Drupal\responsive_image\ResponsiveImageStyleInterface $responsive_image_style
   *   The responsive image style.
   * @param \Drupal\file\FileInterface $file
   *   The file to process.
   * @param bool $inline
   *   Should this thumb written as base64 string.
   *
   * @return array
   *   The attributes array.
   */
  public function buildImageAttributes(ResponsiveImageStyleInterface $responsive_image_style, FileInterface $file, bool $inline = FALSE): array {
    $uri = $file->getFileUri();
    $attributes = [
      'data-src' => JSON::encode($this->buildStyles($uri, $responsive_image_style)),
      'fallback' => $this->buildUrl($uri, $responsive_image_style, 'fallback'),
      'src' => $this->buildUrl($uri, $responsive_image_style, 'preload'),
    ];

    /** @var \Drupal\image\Entity\ImageStyle $preload_style */
    if ($inline === TRUE && ($preload_style = $this->entityTypeManager->getStorage('image_style')->load($responsive_image_style->getPreloadImageStyle()))) {
      $attributes = array_merge($attributes, $this->createBase64Attributes($preload_style, $file));
    }

    return $attributes;
  }

  /**
   * Builds attributes based on an image item.
   *
   * @param \Drupal\image\Plugin\Field\FieldType\ImageItem $item
   *   The image item.
   * @param $allowed_properties
   *   Allowed properties. All other properties are getting ignored from the
   *   item.
   *
   * @return array
   *   An attributes array.
   */
  public function buildImageItemAttributes(ImageItem $item, $allowed_properties = ['alt', 'title']): array {
    $attributes = $item->getValue();
    return array_filter($attributes,
      function ($key) use ($allowed_properties, $attributes) {
        return in_array($key, $allowed_properties) && !empty($attributes[$key]);
      }, ARRAY_FILTER_USE_KEY
    );
  }

  /**
   * Create base64 attributes based on an image and an image_style.
   *
   * @param \Drupal\image\Entity\ImageStyle $image_style
   *   The image style to use.
   * @param \Drupal\file\FileInterface $image
   *   The image to process.
   */
  protected function createBase64Attributes(ImageStyle $image_style, FileInterface $image): array {
    // Make sure to not give access through the base64 if not granted.
    if (!$image->access('view')) {
      return [];
    }

    $attributes = [];
    $image_uri = $image->getFileUri();
    $image_type = $image->getMimeType();
    $derivative_uri = $image_style->buildUri($image_uri);
    if (!file_exists($derivative_uri)) {
      $image_style->createDerivative($image_uri, $derivative_uri);
    }
    $absolute_path = $this->fileSystem->realpath($derivative_uri);

    // If stage_file_proxy is used, omit file_exists check, as it will be
    // downloaded and therefore existing.
    $valid_file = FALSE;
    if ($this->moduleHandler->moduleExists('stage_file_proxy')) {
      $valid_file = TRUE;
    }
    elseif (file_exists($absolute_path)) {
      $valid_file = TRUE;
    }

    if ($absolute_path && $valid_file) {
      $image_file = file_get_contents($absolute_path);
      $base_64_image = base64_encode($image_file);
      $attributes['src'] = "data:$image_type;base64,$base_64_image";

      $image_media = $this->imageFactory->get($absolute_path);
      if ($image_media->isValid()) {
        $attributes['width'] = $image_media->getWidth();
        $attributes['height'] = $image_media->getHeight();
      }
    }

    return $attributes;
  }

  /**
   * @param string $uri
   * @param ResponsiveImageStyleInterface $responsive_image_style
   * @param string $type
   * @return string
   */
  protected function buildUrl(string $uri, ResponsiveImageStyleInterface $responsive_image_style, string $type): string {
    if ($type == 'fallback') {
      if ($responsive_image_style->getFallbackImageStyle() == 'original') {
        return $this->fileUrlGenerator->generateAbsoluteString($uri);
      }
      /** @var \Drupal\image\Entity\ImageStyle $image_style */
      $image_style = $this->entityTypeManager->getStorage('image_style')->load($responsive_image_style->getFallbackImageStyle());
    }

    if ($type == 'preload') {
      /** @var \Drupal\image\Entity\ImageStyle $image_style */
      $image_style = $this->entityTypeManager->getStorage('image_style')->load($responsive_image_style->getPreloadImageStyle());
    }

    if (!$image_style) {
      return $this->fileUrlGenerator->generateAbsoluteString($uri);
    }

    return $image_style->buildUrl($uri);
  }

  /**
   * Build all the styles for an uri based on the responsive_image_style.
   *
   * @param string $uri
   *   The file uri to build the styles with.
   * @param \Drupal\responsive_image\ResponsiveImageStyleInterface $responsive_image_style
   *   The responsive image style.
   *
   * @return array
   *   An array of styles.
   */
  protected function buildStyles(string $uri, ResponsiveImageStyleInterface $responsive_image_style): array {
    $image = $this->imageFactory->get($uri);
    $styles = [];

    /** @var \Drupal\image\ImageStyleInterface $image_style */
    foreach ($this->entityTypeManager->getStorage('image_style')->loadMultiple($responsive_image_style->getImageStyleIds()) as $image_style) {
      if ($image->isValid()) {
        $uri = $image->getSource();
        $dimensions = [
          'width' => $image->getWidth(),
          'height' => $image->getHeight(),
        ];
        $image_style->transformDimensions($dimensions, $uri);
        if ($dimensions['width'] > 0 && !isset($styles[(string)$dimensions['width']])) {
          $styles[(string)$dimensions['width']] = $image_style->buildUrl($uri);
        }
      }
      // Try getting the information from the image style as a fallback.
      else {
        $effects = $image_style->getEffects();
        foreach ($effects as $effect) {
          if (is_a($effect, 'Drupal\image\Plugin\ImageEffect\ResizeImageEffect')) {
            $width = $effect->getConfiguration()['data']['width'];
            if ($width !== NULL && $width > 0) {
              $styles[(string)$width] = $image_style->buildUrl($uri);
            }
          }
        }
      }
    }

    ksort($styles);
    return $styles;
  }

}
